# Hueckel Molecular Orbital Calculation
#     Programmed by Hiroshi Kihara

using LinearAlgebra
using Printf

function definefiles()
    fname = "Butadiene"
    infile = fname * ".dat"
    outfile = fname * ".out"
    return infile, outfile
end

function readdata(infile, outfile)
    f = open(infile, "r")
    fw = open(outfile, "w")

    title = readline(f)
    na, nb, np, nh = parse.(Int64, split(readline(f)))
    charge = na - np
    println(fw, title)
    println(fw, "$na atoms  charge= $charge")
    println(fw)
    H = zeros(na, na)

    for i = 1:nb
        j, k = parse.(Int64, split(readline(f)))
        H[j, k] = 1.0
        H[k, j] = 1.0
    end

    if nh > 0
        for i = 1:nh
            s1, s2, s3 = split(readline(f))
            j = parse(Int64, s1)
            k = parse(Int64, s2)
            hjk = parse(Float64, s3)
            H[j, k] = hjk
            H[k, j] = hjk
        end
    end

    close(f)
    return fw, na, np, H
end

function diag(H)
    """
    Diagonalize using LinearAlgebra.eigen to find eigenvalues and eigenvectors

    return value:
        E : eigen values in descending order
        v : eigen vectors, each vector as a row vector 
        [ [c11, c12, ..., c1n], [c21, c22, ..., c2n], ...  [cn1, cn2, ..., cnn]]
    """
    values, vec = eigen(H)

    E = reverse(values)
    v = reverse(vec, dims=2)

    return E, v

end

function eig_print(fw, n_atom, E, v)
    column = 8
    first = 1
    last = first + column - 1

    println(fw, "  --  Eigenvalues and Eigenvector   --")

    while first <= n_atom
        println(fw)
        if last > n_atom 
            last = n_atom
        end

        for i in first:last
            @printf(fw,"%10d", i)
        end
        println(fw)
        print(fw,"  ")
        for i in first:last
            @printf(fw,"%10.4f", E[i])
        end
        println(fw)
        println(fw)

        for i in 1:n_atom
            @printf(fw,"%2d", i)
            for j in first:last
                @printf(fw,"%10.4f", v[i,j])
            end
            println(fw)
        end
        println(fw)

        first = last + 1
        last = first + column - 1
    end
end

# Calculation of Bond Order Matrix
function calc_pmat(E, v, n, n_elec)

    odd = false
    degenerate = false
    EPS = 0.000001 

    nocc = n_elec ÷ 2      # number of fully occupied orbitals
    if n_elec % 2 != 0
        odd = true
    end
    # Compare the energies of HOMO and the next higher MO, 
    # if smaller than the threshold, considered to be degenerate.
    if (E[nocc] - E[nocc+1]) < EPS 
        degenerate = true
    end
    
    p = zeros(n, n)

    for i in 1:n
        for j in 1:i
            for k in 1:nocc-1
                p[i,j] += 2.0 * v[i,k] * v[j,k]
            end
            if degenerate
                if odd
                    p[i,j] += 0.5 * v[i,nocc]* v[j,nocc] + v[i,nocc+1] * v[j,nocc+1]
                else
                    p[i,j] += v[i,nocc] * v[j,nocc] + v[i,nocc+1] * v[j,nocc+1]
                end
            else
                if odd
                    p[i,j] += v[i,nocc] * v[j,nocc]
                else
                    p[i,j] += 2.0 * v[i,nocc] * v[j,nocc]
                end
            end
        end
    end
    return p
end

function pmat_print(fw, n, p)
    column = 8

    println(fw,"  -- Bond Order Matrix --") 

    first = 1
    last = first + column - 1

    while first <= n
        println(fw)
        if last > n 
            last = n
        end
        for i in first:last
            @printf(fw,"%10d", i)
        end
        println(fw)
        
        for i in 1:n
            @printf(fw,"%2d", i)
            for j in first:last
                @printf(fw,"%10.4f", p[i,j])
            end
            println(fw)
        end

        println(fw)
        first = last + 1
        last = first + column - 1
    end
    close(fw)
end

#  =========  Main program ==========

cd("D:\\Julia")        # Working directory

infile, outfile = definefiles()

fw, na, np, H = readdata(infile, outfile)

E, v = diag(H)

eig_print(fw, na, E, v)

p = calc_pmat(E, v, na, np)

pmat_print(fw, na, p)
